# -*- coding: utf-8 -*-
#
# cwinpy documentation build configuration file (copied from that for bilby)
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import cweqgen
from cweqgen.sphinx import (
    generate_equations_doc,
    generate_docstringexample_doc,
)

sys.path.insert(0, os.path.abspath("../cweqgen/"))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.intersphinx",
    "nbsphinx",
    "sphinxcontrib.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "recommonmark",
    "sphinx_panels",
    "sphinx.ext.todo",
]

autosummary_generate = True
autodoc_docstring_signature = True
autodoc_default_options = {"members-order": "groupwise", "show-inheritance": True}
todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = {
    ".rst": "restructuredtext",
}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = u"cweqgen"
copyright = u"2022, Matthew Pitkin"
author = u"Matthew Pitkin"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
fullversion = cweqgen.__version__

# The short X.Y version.
version = ".".join(fullversion.split(".")[:2])

# The full version, including alpha/beta/rc tags.
release = fullversion

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "__pycache__",
]

add_function_parentheses = True
add_module_names = False
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "cweqgendoc"

html_theme_options = {
    # "style_nav_header_background": "#e3e2e1",
    # "logo_only": True,
    "icon_links": [
        {
            "name": "GitLab",
            "url": "https://git.ligo.org/CW/software/CWEquationGenerator",
            "icon": "fab fa-gitlab",
        },
    ],
    "use_edit_page_button": True,
}

html_context = {
    "gitlab_url": "https://git.ligo.org",
    "gitlab_user": "CW/software",
    "gitlab_repo": "CWEquationGenerator",
    "gitlab_version": "main",
    "doc_path": "docs",
}

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "cweqgen.tex", u"cweqgen Documentation", u"Matthew Pitkin", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "cweqgen", u"cweqgen Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "cweqgen",
        u"cweqgen Documentation",
        author,
        "cwinpy",
        "One line description of project.",
        "Miscellaneous",
    ),
]

numpydoc_show_class_members = False

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "astropy": ("https://docs.astropy.org/en/stable/", None),
    "sympy": ("https://docs.sympy.org/latest/", None),
    "matplotlib": ("https://matplotlib.org/", None),
}


# try and get dictionaries to display better https://stackoverflow.com/a/65195854/1862861
from pprint import pformat


def object_description(object) -> str:
    return pformat(object, indent=4, width=60)


from sphinx.util import inspect

inspect.object_description = object_description

generate_equations_doc("equations.rst")
generate_docstringexample_doc("docstringexample.txt")



"""
Functions to convert between equivalent parameters.
"""


def convert_to_rotation_frequency(gwfrequency=None, rotationperiod=None, **kwargs):
    """
    Convert the GW frequency (assumed to be twice the rotation frequency) or
    the rotation period into the rotation frequency.
    """

    if gwfrequency is not None:
        return gwfrequency / 2.0
    elif rotationperiod is not None:
        return 1.0 / rotationperiod
    else:
        raise ValueError("Required conversion parameters are not present")


def convert_to_rotation_fdot(
    gwfrequency=None,
    rotationfrequency=None,
    rotationperiod=None,
    gwfdot=None,
    rotationpdot=None,
    **kwargs,
):
    """
    Convert the GW frequency (assumed to be twice the rotation frequency) or
    the rotation period and GW rotation frequency derivative or rotation
    period derivative into rotation frequency derivative.
    """

    freq = (
        gwfrequency / 2.0
        if gwfrequency is not None
        else (
            (1.0 / rotationperiod) if rotationperiod is not None else rotationfrequency
        )
    )

    if freq is not None:
        if gwfdot is not None:
            fdot = gwfdot / 2.0
        elif rotationpdot is not None:
            fdot = -rotationpdot * freq ** 2
        else:
            fdot = None

    if freq is None or fdot is None:
        raise ValueError("Required conversion parameters are not present")

    return fdot


def convert_to_rotation_period(gwfrequency=None, rotationfrequency=None, **kwargs):
    """
    Convert the GW frequency (assumed to be twice the rotation frequency) or
    the rotation frequency into the rotation period.
    """

    if gwfrequency is not None:
        return 2.0 / gwfrequency
    elif rotationfrequency is not None:
        return 1.0 / rotationfrequency
    else:
        raise ValueError("Required conversion parameters are not present")


def convert_to_rotation_pdot(gwfrequency=None, rotationfrequency=None, rotationperiod=None, rotationfdot=None, **kwargs):
    """
    Convert the GW frequency (assumed to be twice the rotation frequency) or
    the rotation frequency into the rotation period.
    """

    if rotationfdot is None:
        raise ValueError("Required conversion parameters are not present")
    else:
        fdot = rotationfdot

    if rotationperiod is not None:
        freq = 1.0 / rotationperiod
    elif gwfrequency is not None:
        freq = gwfrequency / 2.0
    elif rotationfrequency is not None:
        freq = rotationfrequency
    else:
        raise ValueError("Required conversion parameters are not present")

    return -fdot / freq ** 2



import abc

from copy import deepcopy
from fractions import Fraction
from astropy.units.format.base import Base

import numpy as np
from numpy import pi

import astropy.units as u
from astropy.units.quantity import Quantity
from astropy.constants import G, c

from matplotlib import pyplot as plt

from sympy import (
    Abs,
    Add,
    Eq,
    expand_power_base,
    I,
    Id,
    lambdify,
    latex,
    Mul,
    Pow,
    powsimp,
    solve,
    Symbol,
    symbols,
    sympify,
)
from sympy.core.function import _coeff_isneg

from .definitions import ALLOWED_VARIABLES, EQN_DEFINITIONS, SUPPLEMENTAL_EQUATIONS

#: dictionary of constants
CONSTANTS = {"G": G, "c": c, "pi": pi}


def constfunc(name):
    return CONSTANTS[name]


def equations(equation, **kwargs):
    """
    This function generates a :class:`~cweqgen.equations.EquationBase` class
    holding a requested equation. This should always be used to generate an
    equation rather than using the :class:`~cweqgen.equations.EquationBase`
    class itself.
    """

    if equation.lower() not in list(EQN_DEFINITIONS.keys()) + list(
        SUPPLEMENTAL_EQUATIONS.keys()
    ):
        raise KeyError(f"Equation '{equation}' is not currently defined")

    # set values for required equation
    eqinfo = (
        EQN_DEFINITIONS[equation.lower()]
        if equation.lower() in EQN_DEFINITIONS
        else SUPPLEMENTAL_EQUATIONS[equation.lower()]
    )
    kwargs["equation"] = equation.lower()
    kwargs["equation_variable"] = eqinfo["variable"]
    kwargs["default_fiducial_values"] = eqinfo["default_fiducial_values"]
    kwargs["alternative_variables"] = eqinfo.get("alternative_variables", [])
    kwargs["converters"] = eqinfo.get("converters", {})

    # reference information
    try:
        kwargs["reference_string"] = eqinfo["reference"].get("short", None)
        kwargs["reference_eqno"] = eqinfo["reference"].get("eqno", None)
        kwargs["reference_adsurl"] = eqinfo["reference"].get("adsurl", None)
        kwargs["reference_bibtex"] = eqinfo["reference"].get("bibtex", None)
    except KeyError:
        # no references given
        pass

    kwargs["latex_string"] = eqinfo["latex_string"]
    kwargs["description"] = eqinfo["description"]

    kwargs["rhs_latex_strings"] = {}
    for key in kwargs["default_fiducial_values"]:
        kwargs["rhs_latex_strings"][key] = (
            ALLOWED_VARIABLES[key]["latex_string"] if key in ALLOWED_VARIABLES else key
        )

    # check whether generating an equation from parts of from a chain
    parts = eqinfo.get("parts", None)

    # generate equation
    if parts is not None:
        kwargs["parts"] = parts

    else:
        chain = eqinfo["chain"]

        # get start equation, given by the first item in chain
        try:
            eq = equations(chain[0])
        except Exception as e:
            raise RuntimeError(f"Could not generate first equation in a chain: {e}")

        eqother = None
        for link in chain[1:]:
            # split into parts
            linkparts = link.split()

            if len(linkparts) != 2:
                raise ValueError("chain components must contain 2 values")

            if "equals" == linkparts[0].strip().lower():
                # an equality
                eqother = equations(linkparts[1].strip())
            elif "rearrange" == linkparts[0].strip().lower():
                # rearrange
                varname = linkparts[1].strip().lower()

                if eqother is not None:
                    eq = eq.rearrange(varname, equal=eqother)
                    eqother = None
                else:
                    eq = eq.rearrange(varname)
            elif "substitute" == linkparts[0].strip().lower():
                # substitute
                subeqname = linkparts[1].strip().lower()
                subeq = equations(subeqname)

                eq = subeq.substitute(eq)

        kwargs["parts"] = eq.parts

    eq = EquationBase(**kwargs)

    # update the equation docstring
    try:
        eq.__doc__ = eqinfo["docstring"]
    except KeyError:
        pass

    # return the equation class
    return eq


class EquationBase:
    __metaclass__ = abc.ABCMeta

    def __init__(self, **kwargs):
        """
        Base class for holding equations.
        """

        self.kwargs = deepcopy(kwargs)  # store copy of initial kwargs

        self.equation_name = kwargs.pop("equation")
        self.variable = kwargs.pop("equation_variable")

        # a list of tuples containing parts of equation
        self.parts = kwargs.pop("parts")

        # dictionary to hold default fiducial values
        self.default_fiducial_values = kwargs.pop("default_fiducial_values")

        # list containing additional keyword values that can be used
        self.alternative_variables = kwargs.pop("alternative_variables", [])

        # dictionary of functions to convert from additional values into
        # required values (keyed on the required values)
        self.converters = kwargs.pop("converters", {})

        self.latex_name = kwargs.pop("latex_string")  # lhs of equation
        self.description = kwargs.pop("description")

        # LaTeX strings for RHS of equation
        self.rhs_latex_strings = kwargs.pop("rhs_latex_strings")

        self.values = self.parse_kwargs(**kwargs)

        # get simple format reference
        self.reference_string = kwargs.pop("reference_string", None)

        # equation number in reference
        self.reference_eqno = kwargs.pop("reference_eqno", None)

        # URL of reference in ADS
        self.reference_adsurl = kwargs.pop("reference_adsurl", None)

        # BibTeX for reference (from ADS)
        self.reference_bibtex = kwargs.pop("reference_bibtex", None)

        # make sure Sympy version of equation is generated
        _ = self.sympy_var
        _ = self.constant

    def parse_kwargs(self, **kwargs):
        """
        Get the required values for the specific equation.
        """

        values = {}

        for key in (
            list(self.default_fiducial_values.keys()) + self.alternative_variables
        ):
            value = self.check_for_alias(key, **kwargs)
            if value is not None:
                values[key] = value

        # perform conversions if required
        for key in self.converters:
            if key not in values:
                try:
                    values[key] = self.converters[key](**values)
                except ValueError:
                    pass

        return values

    @staticmethod
    def check_for_alias(key, **kwargs):
        """
        Check whether any alias of the key is given in the keyword arguments.
        """

        # check aliases
        if key in ALLOWED_VARIABLES:
            aliases = ALLOWED_VARIABLES[key]["aliases"]
        else:
            aliases = [key]

        for alias in aliases:
            if alias in kwargs:
                value = kwargs[alias]

                # check value has compatible units
                if key in ALLOWED_VARIABLES:
                    if not isinstance(value, Quantity):
                        if ALLOWED_VARIABLES[key]["units"] is not None:
                            value *= u.Unit(ALLOWED_VARIABLES[key]["units"])
                        else:
                            value = Quantity(value)
                    elif (
                        isinstance(value, Quantity)
                        and ALLOWED_VARIABLES[key]["units"] is not None
                    ):
                        try:
                            _ = value.to(ALLOWED_VARIABLES[key]["units"])
                        except (u.UnitConversionError, ValueError) as e:
                            raise IOError(
                                f"{ALLOWED_VARIABLES[key]['description']} units are not compatible:\n{e}"
                            )

                    # check value has correct sign
                    if ALLOWED_VARIABLES[key]["sign"] is not None:
                        strrep = (
                            "np.array("
                            + np.array2string(np.asarray(value.value), separator=", ")
                            + ")"
                        )

                        if not eval(
                            "np.all(" + strrep + ALLOWED_VARIABLES[key]["sign"] + ")"
                        ):
                            raise ValueError(
                                f"{ALLOWED_VARIABLES[key]['description']} does not have the correct sign"
                            )

                return value

        return None

    def calculate_fiducial(self, **kwargs):
        """
        Calculate and return the product of the fiducial components of the
        equation (excluding the constants) in SI units.

        Keyword arguments can be passed using the keys of the
        :attr:`.EquationBase.default_fiducial_values` giving the values at
        which to perform the calculation. Otherwise the default values, or
        values set at initialisation are used. These can be 1d arrays. If
        the arrays have different lengths, or the "mesh" keyword argument
        is given and is True, then a mesh grid will be created over the
        space and the values returned on that mesh. If arrays of equal
        length are given, and the "mesh" keyword is not given, then values
        will be calculated assuming for each set of equivalently positioned
        values in the array.
        """

        # fiducialunits = 1.0
        funcargs = {}
        arglens = []  # store lengths of arguments

        self._unitdic = {}

        values = self.parse_kwargs(**kwargs)

        for key in self.var_names:
            if key in values:
                # use provided value
                val = values[key]
            else:
                val = self.default_fiducial_values[key]

            if not isinstance(val, Quantity):
                val = Quantity(val)

            funcargs[key] = (
                np.abs(val.si) if ALLOWED_VARIABLES[key]["sign"] == ">= 0" else val.si
            )

            try:
                arglens.append(len(funcargs[key]))
            except TypeError:
                arglens.append(1)

        # check whether multiple values are arrays and create a mesh if
        # necessary
        usemesh = kwargs.get("mesh", False)
        mesh = None
        if sum([length > 1 for length in arglens]) > 1:
            idx = np.argwhere(np.array(arglens) > 1).flatten()

            # create mesh is "mesh" is True or arguments have different lengths
            if usemesh or not np.all(
                (np.array([arglens[i] for i in idx]) - arglens[idx[0]]) == 0
            ):
                keys = [list(funcargs.keys())[i] for i in idx]

                mesh = np.meshgrid(*[funcargs[key] for key in keys])
                for i, key in enumerate(keys):
                    funcargs[key] = mesh[i].flatten()

        # evaluate equation (loop through each part)
        fiducial = 1.0
        for key in funcargs:
            eveq = self._sympy_var_lambda[key](**{key: funcargs[key]})
            # NOTE: can't use *= as it causes a SegFault
            fiducial = fiducial * eveq

        if mesh is not None:
            fiducial = np.reshape(fiducial, mesh[0].shape)

        if isinstance(fiducial, Quantity):
            return fiducial.si.decompose()
        else:
            return fiducial

    def equation(self, displaytype="string", **latexkwargs):
        """
        Generate the LaTeX string for the equation.

        Parameters
        ----------
        displaytype: str
            By default this will return a string containing the LaTeX equation
            text (without bounding "$" symbols). If using a Jupyter Notebook
            this will show as a formatted LaTeX equation. Alternatively, set to
            "matplotlib" to have the output returned as a Matplotlib figure
            object containing the equation.
        latexkwargs: dict
            Keyword parameters that can be passed to the
            :func:`sympy.printing.latex.latex` function. By default the
            ``fold_frac_powers`` option is set to True, the ``root_notation``
            option is set to False and the LaTeX symbol names are defined by
            the equation definition values.
        """

        # use powsimp to put values with common exponents together
        seq_const = powsimp(self.sympy_const, force=True)
        seq_var = powsimp(self.sympy_var, force=True)

        # set defaults
        symrep = {symbols(key): val for key, val in self.rhs_latex_strings.items()}
        symrep[symbols(self.variable)] = self.latex_name

        latexkwargs.setdefault("root_notation", True)
        latexkwargs.setdefault("fold_frac_powers", True)
        latexkwargs.setdefault("long_frac_ratio", 2.0)

        mode = latexkwargs.pop("mode", "plain")
        delim = "$" if displaytype == "matplotlib" or mode == "inline" else ""

        if abs(seq_const) != 1:
            latex_equation_const = latex(seq_const, **latexkwargs)
        else:
            latex_equation_const = (
                ""
                if seq_const == 1 or ALLOWED_VARIABLES[self.variable]["sign"] == ">= 0"
                else "-"
            )

        latexkwargs["root_notation"] = False
        latexkwargs.setdefault("symbol_names", symrep)
        latex_equation_var = latex(seq_var, **latexkwargs)

        latex_equation = f"{delim}{self.latex_name} = {latex_equation_const}{latex_equation_var}{delim}"

        if displaytype.lower() == "matplotlib":
            return EquationLaTeXToImage(latex_equation)
        else:
            return EquationLaTeXString(latex_equation)

    @property
    def eqn(self):
        """
        The equation as a string.
        """

        return self.equation()

    def __str__(self):
        return str(self.equation())

    def __repr__(self):
        return str(self.equation())

    def _repr_latex_(self):
        return "$" + str(self.equation()) + "$"

    def fiducial_equation(self, dp=2, brackets="()", displaytype="string", **kwargs):
        """
        Generate the LaTeX string for the equation inserting in fiducial values.

        Parameters
        ----------
        dp: int
            The number of decimal places to use for non-integer fiducial values.
        brackets: str
            The style of brackets to use, e.g., "()", "{}" or "[]". Defaults is
            round parentheses "()".
        displaytype: string
            By default this will return a string containing the LaTeX equation
            text (without bounding "$" symbols). If using a Jupyter Notebook
            this will show as a formatted LaTeX equation. Alternatively, set to
            "matplotlib" to have the output returned as a Matplotlib figure
            object containing the equation.
        """

        latex_equation = self.latex_name + " = "

        # add in coefficient
        values = deepcopy(self.values)
        for key, val in kwargs.items():
            values[key] = val
        values = self.parse_kwargs(**values)

        coeff = self.evaluate(**values)

        if not isinstance(coeff, Quantity):
            # convert into Quantity
            coeff = Quantity(coeff)

        latex_equation += coeff.to_string(precision=(dp + 1), format="latex").replace(
            "$", ""
        )

        if brackets not in ["()", "{}", "[]", None]:
            raise ValueError(f"Bracket type {brackets} is not recognised")

        lbrace = r"\left" + brackets[0] if brackets is not None else ""
        rbrace = r"\right" + brackets[1] if brackets is not None else ""

        fiducial = ""

        for key in self.var_names:
            arg = self._sympy_var_parts[key]
            varlatex = (
                self.rhs_latex_strings[key]
                if not isinstance(arg, Pow)
                else latex(
                    arg.base, symbol_names={symbols(key): self.rhs_latex_strings[key]}
                )
            )

            # get exponent
            exp = 1 if not isinstance(arg, Pow) else Fraction(*arg.exp.as_numer_denom())

            # get value
            if key in values:
                # use provided value
                val = Quantity(values[key])
            else:
                val = Quantity(self.default_fiducial_values[key])

            if exp != 1:
                # evaluate base for cases such as args being (n + 1)^x
                val = (
                    float(arg.base.subs([(symbols(key), val.value)]).evalf()) * val.unit
                )

            if exp < 0:
                numerator = val.to_string(precision=(dp + 1), format="latex").replace(
                    "$", ""
                )
                denominator = varlatex
            else:
                denominator = val.to_string(precision=(dp + 1), format="latex").replace(
                    "$", ""
                )
                numerator = varlatex

            if abs(exp) != 1:
                expstr = "^{" + str(abs(exp)) + "}"
            else:
                expstr = ""

            fiducial += (
                rf"{lbrace}\frac{{{numerator}}}{{{denominator}}}{rbrace}{expstr} "
            )

        latex_equation += fiducial

        if displaytype.lower() == "matplotlib":
            return EquationLaTeXToImage("$" + latex_equation + "$")
        else:
            return EquationLaTeXString(latex_equation)

    def evaluate(self, **kwargs):
        """
        Evaluate the equation using the given values.

        Keyword arguments can be passed using the keys of the
        :attr:`.EquationBase.default_fiducial_values` giving the values at
        which to perform the calculation. Otherwise the default values, or
        values set at initialisation are used. These can be 1d arrays,
        where if more that one value is an array of then a mesh grid will
        be created over the space and the values returned on that mesh.

        Parameters
        ----------
        value: bool
            If value is False (the default) the evaluated equation will be
            returned as an :class:`astropy.units.Quantity` object. If True
            just the values will be returned.
        """

        const = self.constant
        fid = self.calculate_fiducial(**kwargs)

        value = const * fid

        if ALLOWED_VARIABLES[self.variable]["sign"] == ">= 0":
            value = np.abs(value)

        if isinstance(value, Quantity):
            if not kwargs.get("value", False):
                if ALLOWED_VARIABLES[self.variable]["units"] is not None:
                    return value.to(ALLOWED_VARIABLES[self.variable]["units"])
                else:
                    return value.si.decompose()
            else:
                # return as values rather than Quantity object
                return value.si.decompose().value
        else:
            return value

    def __call__(self, **kwargs):
        return self.evaluate(**kwargs)

    def rearrange(self, newvar, fidval=None, equal=None):
        r"""
        Rearrange the equation so that a different variable is on the left hand side,
        i.e., solve the equation for the given variable.

        Parameters
        ----------
        newvar: str
            The variable should be rearranged to the LHS of the equation.
        fidval: float, Quantity
            The value to use for the original LHS value. If not given this
            will be based on the original fiducial values.
        equal: EquationBase
            You can pass another equation to set as equal to the current
            equation before rearranging.

        Returns
        -------
        neweq: EquationBase
            Returns a new equation. The current equation is left unchanged.

        Example
        -------
        For example rearrange the braking index equation to put the
        frequency derivative :math:`\dot{f}` on the left hand side:

        >>> eqn = equations("brakingindex")
        >>> reqn = eqn.rearrange("rotationfdot")

        which gives:

        .. math::

            \dot{f}_{\rm rot} = \frac{1}{n^{1/2}} \left(\ddot{f}_{\rm rot} f_{\rm rot}\right)^{1/2}
        """

        if newvar not in self.var_names:
            raise KeyError(f"{newvar} is not allowed")

        if fidval is None:
            # use current fiducial values to get value of parameter being
            # swapped from LHS
            curval = self.evaluate()
        else:
            curval = fidval

        # check whether equating to other equation
        if isinstance(equal, EquationBase):
            if self.variable != equal.variable:
                raise ValueError(
                    "Equation can only be equated if the lhs variable is the same"
                )
            eq = Eq(equal.sympy.rhs, self.sympy.rhs)
        else:
            eq = self.sympy

        # rearrange using solve (use the last value in solution in case two solutions
        # from sqrt). Expand out any variables with the same powers, so that they form
        # separate parts of the new equation
        neweq = expand_power_base(solve(eq, symbols(newvar))[-1], force=True)

        newkwargs = deepcopy(self.kwargs)

        # set new equation parts
        newkwargs["parts"] = self.generate_parts(neweq)

        newfiducial = newkwargs.pop("default_fiducial_values")
        newfiducial.pop(newvar)
        newfiducial[self.equation_name] = curval

        if isinstance(equal, EquationBase):
            newfiducial.update(equal.kwargs["default_fiducial_values"])
        newkwargs["default_fiducial_values"] = newfiducial

        newkwargs["latex_string"] = ALLOWED_VARIABLES[newvar]["latex_string"]
        newkwargs["description"] = ALLOWED_VARIABLES[newvar]["description"]

        newkwargs["rhs_latex_strings"].pop(newvar)

        if not isinstance(equal, EquationBase):
            newkwargs["rhs_latex_strings"][self.variable] = self.latex_name
        else:
            newkwargs["rhs_latex_strings"].update(equal.kwargs["rhs_latex_strings"])

        newkwargs["equation"] = newvar  # reset the equation name
        newkwargs["equation_variable"] = newvar

        # create new EquationBase with updated constants and default fiducial values
        return EquationBase(**newkwargs)

    def substitute(self, other):
        r"""
        Substitute another equation into the current equation.

        Parameters
        ----------
        other: EquationBase
            The other equation to substitute into the current equation.

        Returns
        -------
        neweq: EquationBase
            Returns a new equation. The current equation is left unchanged.

        Example
        -------
        For example put the :math:`h_0` spin-down limit in terms of the
        braking index :math:`n` and second frequency derivative
        :math:`\ddot{f}` (with the help of
        :meth:`~cweqgen.equations.EquationBase.rearrange`):

        >>> # get braking index equation
        >>> eqn = equations("brakingindex")
        >>> # rearrange to put fdot on lhs
        >>> reqn = eqn.rearrange("rotationfdot")
        >>> # get h0 spindown equation
        >>> eqnh0sd = equations("h0spindown")
        >>> # substitute in the rearranged equation
        >>> neweq = eqnh0sd.substitute(reqn)

        with gives:

        .. math::

            h_0^{\rm sd} = \frac{\sqrt{10} \sqrt{G}}{2 c^{3/2}}\frac{I_{zz}^{1/2} \ddot{f}_{\rm rot}^{1/4}}{d \left(n f_{\rm rot}\right)^{1/4}}
        """

        if not isinstance(other, EquationBase):
            raise TypeError("Other equation is not the right type")

        try:
            rhs = self.sympy.rhs
            neweq = rhs.subs([(other.sympy.lhs, other.sympy.rhs)])
        except Exception as e:
            raise RuntimeError(f"Could not perform substitution: {e}")

        newkwargs = deepcopy(self.kwargs)
        newfiducial = newkwargs.pop("default_fiducial_values")
        newfiducial.pop(other.variable)
        newfiducial.update(other.default_fiducial_values)

        newkwargs["default_fiducial_values"] = newfiducial

        newkwargs["rhs_latex_strings"].pop(other.variable)
        newkwargs["rhs_latex_strings"].update(other.rhs_latex_strings)

        newkwargs["parts"] = self.generate_parts(expand_power_base(neweq, force=True))

        return EquationBase(**newkwargs)

    def __str__(self):
        return str(self.equation())

    @property
    def sympy_const(self):
        """
        Construct and return a :class:`sympy.core.mul.Mul` containing the
        constants in the equation.
        """

        if not hasattr(self, "_sympy_const"):
            self._sympy_const = 1
            sympy_const_unit_values = {}

            for arg in self.sympy.rhs.args:
                if arg.is_constant():
                    self._sympy_const *= arg
                else:
                    if isinstance(arg, Symbol):
                        name = arg.name
                    elif isinstance(arg, Pow):
                        name = str(arg.base)
                    else:
                        name = None

                    if name in CONSTANTS:
                        sympy_const_unit_values[name] = CONSTANTS[name]

                        if ALLOWED_VARIABLES[self.variable]["sign"] == ">= 0":
                            # make sure values are positive
                            if isinstance(arg, Pow):
                                if _coeff_isneg(arg.args[0]):
                                    arg = (-1 * arg.args[0]) ** arg.args[1]
                            else:
                                if _coeff_isneg(arg):
                                    arg = -1 * arg

                        self._sympy_const *= arg

            # evaluate constant by creating a lamdified function
            if self._sympy_const != 1:
                # make sure constant isn't imaginary
                if any([arg == I for arg in self._sympy_const.args]):
                    self._sympy_const = self._sympy_const.replace(I, 1)

                constf = lambdify(
                    [symbols(name) for name in sympy_const_unit_values.keys()],
                    self._sympy_const,
                    modules=["numpy"],
                )
                constant = constf(**sympy_const_unit_values)

                if isinstance(constant, Quantity):
                    self._constant = constant.si.decompose()
                else:
                    self._constant = constant
            else:
                self._constant = self._sympy_const

        return self._sympy_const

    @property
    def constant(self):
        """
        Return the constant coefficient factor in the equation in SI units
        (if it has dimensions).
        """

        if not hasattr(self, "_constant"):
            _ = self.sympy_const

        return self._constant

    @property
    def sympy_var(self):
        """
        Construct and return a :class:`sympy.core.mul.Mul` containing the
        variables in the equation.
        """

        if not hasattr(self, "_sympy_var"):
            self._sympy_var = 1
            self._sympy_var_parts = {}
            self._sympy_var_lambda = {}

            # get variable names
            self._var_names = self._gather_var_argnames(self.sympy.rhs)
            self._gather_var_parts(self.sympy.rhs.args)

        return self._sympy_var

    @staticmethod
    def _gather_var_argnames(eqn):
        """
        Find the variable names in an equation.
        """

        argnames = [arg.name for arg in eqn.atoms() if isinstance(arg, Symbol)]
        for arg in list(argnames):
            if arg in CONSTANTS:
                argnames.remove(arg)

        return argnames

    def _gather_var_parts(self, args):
        """
        Recursively get all the separate parts of the equation
        """

        for arg in args:
            if not arg.is_constant():
                if isinstance(arg, Symbol):
                    name = arg.name
                elif isinstance(arg, (Abs, Add, Pow)):
                    name = self._gather_var_argnames(arg)
                    if len(name) > 1:
                        # recursively gather parts
                        if isinstance(arg, Abs):
                            abseqn = expand_power_base(arg.args[0], force=True).args
                        else:
                            abseqn = expand_power_base(arg, force=True).args

                        self._gather_var_parts(abseqn)
                        continue
                    elif len(name) == 1:
                        name = name[0]
                    else:
                        name = str(arg.base)
                else:
                    raise TypeError("Value must be a Symbol, Pow, Add or Abs")

                if name not in CONSTANTS:
                    # set variables to abs if required
                    if (
                        ALLOWED_VARIABLES[name]["sign"] is None
                        and ALLOWED_VARIABLES[self.variable]["sign"] == ">= 0"
                    ):
                        if isinstance(arg, Pow):
                            arg = Abs(arg.args[0]) ** arg.args[1]
                        else:
                            arg = Abs(arg)

                    self._sympy_var *= arg

                    # store each part
                    self._sympy_var_parts[name] = arg

                    # create functions for each part of the equation
                    self._sympy_var_lambda[name] = lambdify(
                        [symbols(name)], arg, modules=["numpy"]
                    )

    @property
    def var_names(self):
        """
        Get variable names from Sympy representation.
        """

        if not hasattr(self, "_var_names"):
            _ = self.sympy_var

        return self._var_names

    @property
    def sympy_var_func(self):
        if not hasattr(self, "_sympy_var_func"):
            _ = self.sympy_var

        return self._sympy_var_func

    @property
    def sympy(self):
        """
        Construct and return the equation as a
        :class:`sympy.core.relational.Equality`.
        """

        if not hasattr(self, "_sympy"):
            eqstr = " * ".join([f"({c[0]})**({c[1]})" for c in self.parts])
            sympeq = sympify(eqstr)
            self._sympy = Eq(symbols(self.variable), sympeq)

        return self._sympy

    @staticmethod
    def generate_parts(eqn, pow=1):
        """
        Generate a list of "parts" from a Sympy equation.
        """

        parts = []

        for arg in eqn.args:
            if isinstance(arg, (Abs, Mul)):
                parts.extend(EquationBase.generate_parts(arg, pow=pow))
            elif isinstance(arg, Pow):
                if isinstance(arg.base, (Abs, Mul, Pow)):
                    exp = pow * arg.exp
                    parts.extend(EquationBase.generate_parts(arg.base, pow=exp))
                else:
                    parts.append((str(arg.base), str(arg.exp)))
            elif isinstance(arg, Symbol):
                if pow != 1 and len(eqn.args) == 2:
                    # it's got to this part from a Pow object
                    parts.append((arg.name, str(pow * eqn.args[-1])))
                    break
                else:
                    parts.append((arg.name, str(pow)))
            elif str(arg) != "1":
                parts.append((str(arg), "1"))

        return parts

    @staticmethod
    def _frequency_converter(starteqn, end):
        """
        Given an equation, and a required parameter, step through a set of
        frequency conversions until the equation is parameterised with the
        required variables.

        Parameters
        ----------
        starteqn: EquationBase
            The original equation
        end: str
            The variable name for the parameter that you want to equation to
            contain after the conversions are performed.

        Returns
        -------
        neweqn: EquationBase
            A new :class:`~cweqgen.equations.EquationBase` equation using the
            requested variable to parameterise it.
        """

        # the conversion for frequency equations to loop through
        chain = [
            ("rotationperiod", equations("rotationperiod_to_angularrotationfrequency")),
            (
                "angularrotationfrequency",
                equations("angularrotationfrequency_to_angulargwfrequency"),
            ),
            ("angulargwfrequency", equations("angulargwfrequency_to_gwfrequency")),
            ("gwfrequency", equations("gwfrequency_to_rotationfrequency")),
            ("rotationfrequency", equations("rotationfrequency_to_period")),
        ]

        # find the index of the end point
        endidx = ([link[0] for link in chain]).index(end)

        neweqn = None
        # find the starting conversion equation
        for i in range(len(chain)):
            if chain[i][0] in starteqn.var_names:
                break

        while True:
            # loop over conversions until finished
            if neweqn is None:
                neweqn = starteqn.substitute(chain[i][1])
            else:
                neweqn = neweqn.substitute(chain[i][1])

            i = (i + 1) % len(chain)
            if i == endidx:
                break

        # the conversion for frequency derivative equations to loop through
        chaindot = [
            (
                "rotationperiod",
                "rotationpdot",
                [equations("rotationpdot_to_angularrotationfdot"), chain[0][1]],
            ),
            (
                "angularrotationfrequency",
                "angularrotationfdot",
                [equations("angularrotationfdot_to_angulargwfdot"), chain[1][1]],
            ),
            (
                "angulargwfrequency",
                "angulargwfdot",
                [equations("angulargwfdot_to_gwfdot"), chain[2][1]],
            ),
            (
                "gwfrequency",
                "gwfdot",
                [equations("gwfdot_to_rotationfdot"), chain[3][1]],
            ),
            (
                "rotationfrequency",
                "rotationfdot",
                [equations("rotationfdot_to_period"), chain[4][1]],
            ),
        ]

        # find the index of the end point
        endidxdot = ([link[0] for link in chaindot]).index(end)

        # find the starting conversion equation
        for i in range(len(chaindot)):
            if chaindot[i][1] in starteqn.var_names:
                break

        while True:
            # loop over conversions until finished
            for eqn in chaindot[i][2]:
                if neweqn is None and eqn.variable in starteqn.var_names:
                    neweqn = starteqn.substitute(eqn)
                elif eqn.variable in neweqn.var_names:
                    neweqn = neweqn.substitute(eqn)

            i = (i + 1) % len(chaindot)
            if i == endidxdot:
                break

        return neweqn

    def to(self, newvariable):
        """
        Return a new version of the equation in terms of a new variable.
        Currently this is designed to convert between frequency-like and
        frequency first derivative parameters (it does not do higher frequency
        derivatives), e.g., converting an equation from using frequency and
        frequency derivative, to one using period and period derivative.

        Parameters
        ----------
        newvariable: str
            The new parameterisation of the equation. This can be one of:
            "rotationfrequency", "rotationperiod", "gwfrequency",
            "angularrotationfrequency", or "angulargwfrequency" (or any of the
            allowed aliases for these in
            :obj:`~cwinpy.definition.ALLOWED_VARIABLES`). If the equation
            contains equivalents of the variable and/or their first frequency
            derivative then both will be converted.

        Returns
        -------
        neweqn: EquationBase
            The equation in the new parameterisation.
        """

        # find the variable name
        for var in ALLOWED_VARIABLES:
            if newvariable in ALLOWED_VARIABLES[var]["aliases"]:
                break
        else:
            raise ValueError(f"{newvariable} is not a recognised variable name.")

        if var not in [
            "rotationfrequency",
            "rotationperiod",
            "gwfrequency",
            "angularrotationfrequency",
            "angulargwfrequency",
        ]:
            raise ValueError(f"No conversion is currently available for {var}")

        return self._frequency_converter(self, var)


class EquationLaTeXString:
    def __init__(self, latexstring):
        """
        Class to hold a LaTeX equation string. It has a _repr_latex_ method to
        hook into the Jupyter notebook rich display system
        https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
        and show the resulting string as a LaTeX equation.

        Parameters
        ----------
        latexstring: str
            The LaTeX string defining the equation.
        """

        self.text = str(latexstring)

    def __repr__(self):
        return self.text

    def __str__(self):
        return self.text

    def _repr_latex_(self):
        return "$" + self.text + "$"


class EquationLaTeXToImage:
    def __init__(self, latexstring, dpi=200):
        """
        Class to hold a LaTeX equation string and covert it to an image for
        display in a Jupyter notebook.

        Parameters
        ----------
        latexstring: str
            The LaTeX string defining the equation.
        dpi: int
            The resolution (dots per inch) of the output plot.
        """

        self.text = str(latexstring)

        self.dpi = dpi
        self.fig, self.ax = plt.subplots(1)

        try:
            t = self.ax.text(0.05, 0.5, self.text, usetex=True)
        except RuntimeError:
            t = self.ax.text(0.0, 0.5, self.text)
        self.ax.axis("off")

        # crop figure to tight around the text
        bb = t.get_window_extent(renderer=self.fig.canvas.get_renderer())
        transf = self.ax.transData.inverted()
        bb_datacoords = bb.transformed(transf)

        rect = bb_datacoords.get_points().flatten()
        rect[3] += 0.1 * rect[3]  # add 10% on to upper value for some reason!

        tight_params = {"rect": rect}

        self.fig.set_dpi(self.dpi)
        self.fig.set_tight_layout(tight_params)

    def savefig(self, *arg, **kwargs):
        kwargs.setdefault("dpi", self.dpi)
        return self.fig.savefig(*arg, **kwargs)



from ..equations import equations
from ..definitions import EQN_DEFINITIONS


def generate_equations_doc(docfile):
    """
    Helper function to automatically generate a documentation page containing
    all the available equations within cweqgen.

    Parameters
    ----------
    docfile: str:
        The output file for the documentation.
    """

    doccontents = """
#########
Equations
#########

The currently implemented equations are:

"""

    references = """\
References
----------
"""

    usedreferences = []
    usedrefurls = []
    refcount = 1

    for eqn in EQN_DEFINITIONS:
        eqstr = ""

        # create equation
        eq = equations(eqn)

        if eq.reference_string in usedreferences:
            refnum = usedreferences.index(eq.reference_string) + 1
        else:
            usedreferences.append(eq.reference_string)
            usedrefurls.append(eq.reference_adsurl)
            refnum = refcount
            refcount += 1

        eqstr += """\
{0}
{1}
""".format(
            eq.description, "-" * len(eq.description)
        )

        eqstr += """
This equation can be accessed from the :func:`~cweqgen.equations.equations` function using
the name ``{}``.
""".format(
            eqn
        )

        eqno = "" if eq.reference_eqno is None else f"Eqn. {eq.reference_eqno} in "
        eqstr += """
The generated equation ({0} [{1}]_) is:

.. math::

    {2}
""".format(
            eqno, refnum, eq.equation()
        )

        eqstr += """
The fiducial values defined for this equation are:

.. math::

    {}
""".format(
            eq.fiducial_equation()
        )

        eqstr += """
.. note::

    These fiducial values are just those defined within this package and may not be representative
    of fiducial values used elsewhere in the literature.
"""

        eqstr += """
To generate the equation as calculated at particular values, the
:func:`~cweqgen.equations.equations` can be used as

.. py:function:: equations("{0}", {1})
    :noindex:
""".format(
            eq.equation_name,
            ", ".join(
                [
                    "{}={}".format(fid, str(val))
                    for fid, val in eq.default_fiducial_values.items()
                ]
            ),
        )

        # add doc string lines
        for line in eq.__doc__.split("\n"):
            eqstr += f"    {line}\n"

        doccontents += eqstr + "\n"

    # add in list of references
    for i in range(len(usedreferences)):
        references += """
.. [{0}] {1} [`ADS URL <{2}>`_]
""".format(
            (i + 1), usedreferences[i], usedrefurls[i]
        )

    with open(docfile, "w") as fp:
        fp.write(doccontents + references)


def generate_docstringexample_doc(docfile, eqn="h0"):
    """
    Helper function to output an example docstring.

    Parameters
    ----------
    docfile: str
        The output file for the documentation.
    eqn: str
        The name of the equation for which the docstring is required.
    """

    # write out example docstring
    with open(docfile, "w") as fp:
        fp.write(EQN_DEFINITIONS[eqn]["docstring"])



import astropy.units as u
from astropy.units.quantity import Quantity
from numpy import pi

from .converters import *


#: allowed variables for equations
ALLOWED_VARIABLES = {
    "h0": {
        "description": "Gravitational wave amplitude",
        "latex_string": r"h_0",
        "aliases": ["h0", "h_0"],
        "units": None,
        "sign": ">= 0",
    },
    "ellipticity": {
        "description": "Neutron star ellipticity",
        "latex_string": r"\varepsilon",
        "aliases": [
            "ellipticity",
            "ell",
            "eps",
            "epsilon",
            "𝜀",
        ],
        "units": None,
        "sign": ">= 0",
    },
    "massquadrupole": {
        "description": "Mass quadrupole moment (l=m=2)",
        "latex_string": r"Q_{22}",
        "aliases": ["massquadrupole", "q22", "q_22"],
        "units": "kg m^2",
        "sign": ">= 0",
    },
    "rotationfrequency": {
        "description": "Source rotational frequency",
        "latex_string": r"f_{\rm rot}",
        "aliases": [
            "rotationfrequency",
            "frot",
            "spinfrequency",
            "fspin",
            "f0rot",
            "f0spin",
        ],
        "units": "Hz",
        "sign": ">= 0",
    },
    "angularrotationfrequency": {
        "description": "Source angular rotational frequency",
        "latex_string": r"\Omega_{\rm rot}",
        "aliases": [
            "angularrotationfrequency",
            "omegarot",
            "omrot",
            "omega0rot",
            "om0rot",
            "Ω",
            "Ωrot",
            "Ω0rot",
        ],
        "units": "rad / s",
        "sign": ">= 0",
    },
    "gwfrequency": {
        "description": "Gravitational-wave frequency",
        "latex_string": r"f_{\rm gw}",
        "aliases": ["gwfrequency", "fgw", "f0gw"],
        "units": "Hz",
        "sign": ">= 0",
    },
    "angulargwfrequency": {
        "description": "Angular gravitational-wave frequency",
        "latex_string": r"\Omega_{\rm gw}",
        "aliases": ["angulargwfrequency", "omegagw", "omgw", "omega0gw", "om0gw", "Ωgw", "Ω0gw"],
        "units": "rad / s",
        "sign": ">= 0",
    },
    "rotationperiod": {
        "description": "Source rotational period",
        "latex_string": r"P",
        "aliases": ["rotationperiod", "prot", "p0rot"],
        "units": "s",
        "sign": ">= 0",
    },
    "rotationfdot": {
        "description": "Source rotational frequency derivative",
        "latex_string": r"\dot{f}_{\rm rot}",
        "aliases": ["rotationfdot", "frotdot", "f1rot", "f1spin"],
        "units": "Hz / s",
        "sign": None,
    },
    "angularrotationfdot": {
        "description": "Source angular rotational frequency derivative",
        "latex_string": r"\dot{\Omega}_{\rm rot}",
        "aliases": ["angularrotationfdot", "omrotdot", "om1rot", "om1spin"],
        "units": "rad / s^2",
        "sign": None,
    },
    "gwfdot": {
        "description": "Gravitational-wave frequency derivative",
        "latex_string": r"\dot{f}_{\rm gw}",
        "aliases": ["gwfdot", "fdotgw", "f1gw"],
        "units": "Hz / s",
        "sign": None,
    },
    "angulargwfdot": {
        "description": "Gravitational-wave angular frequency derivative",
        "latex_string": r"\dot{\Omega}_{\rm gw}",
        "aliases": ["angulargwfdot", "omgwdot", "om1gw"],
        "units": "rad / s^2",
        "sign": None,
    },
    "rotationpdot": {
        "description": "Source rotational period derivative",
        "latex_string": r"\dot{P}",
        "aliases": ["pdot", "p0dot", "rotationpdot"],
        "units": "s / s",
        "sign": None,
    },
    "rotationfddot": {
        "description": "Source rotational frequency second derivative",
        "latex_string": r"\ddot{f}_{\rm rot}",
        "aliases": ["rotationfddot", "frotddot", "f2rot", "f2spin"],
        "units": "Hz / s / s",
        "sign": None,
    },
    "gwfdot": {
        "description": "Gravitational-wave second frequency derivative",
        "latex_string": r"\dot{f}_{\rm gw}",
        "aliases": ["gwfddot", "fddotgw", "f2gw"],
        "units": "Hz / s / s",
        "sign": None,
    },
    "momentofinertia": {
        "description": "Principal moment of inertia about the rotation axis",
        "latex_string": r"I_{zz}",
        "aliases": ["momentofinertia", "izz", "i38"],
        "units": "kg m^2",
        "sign": ">= 0",
    },
    "distance": {
        "description": "Distance to the source",
        "latex_string": "d",
        "aliases": ["distance", "d", "r"],
        "units": "kpc",
        "sign": ">= 0",
    },
    "brakingindex": {
        "description": "The braking index of a pulsar",
        "latex_string": "n",
        "aliases": ["brakingindex", "n"],
        "units": None,
        "sign": None,
    },
    "characteristicage": {
        "description": "The characteristic age of a pulsar",
        "latex_string": r"\tau",
        "aliases": ["characteristicage", "tau", "𝜏"],
        "units": "yr",
        "sign": ">= 0",
    },
    "luminosity": {
        "description": "The luminosity of a source",
        "latex_string": "L",
        "aliases": [
            "luminosity",
            "l",
            "spindownluminosity",
            "gwluminosity",
            "lgw",
            "lsd",
        ],
        "units": "W",
        "sign": ">= 0",
    },
}


class EqDict(dict):
    """
    Dictionary class to hold equation definitions.
    """

    def __setitem__(self, key, subdict):
        super(EqDict, self).__setitem__(key, {})

        try:
            self[key]["description"] = subdict["description"]
        except KeyError:
            raise KeyError("Equation dictionary must contain a 'description'")

        try:
            self[key]["variable"] = subdict["variable"]
        except KeyError:
            raise KeyError("Equation dictionary must contain a 'variable'")

        try:
            self[key]["latex_string"] = subdict["latex_string"]
        except KeyError:
            raise KeyError("Equation dictionary must contain a 'latex_string'")

        try:
            self[key]["default_fiducial_values"] = subdict["default_fiducial_values"]
        except KeyError:
            raise KeyError("Equation dictionary must contain 'default_fiducial_values'")

        try:
            self[key]["parts"] = subdict["parts"]
        except KeyError:
            try:
                self[key]["chain"] = subdict["chain"]
            except KeyError:
                raise KeyError("Equation dictionary must contain 'parts' or 'chain'")

        self[key]["alternative_variables"] = subdict.get("alternative_variables", [])
        self[key]["converters"] = subdict.get("converters", {})

        if "reference" in subdict:
            self[key]["reference"] = subdict.get("reference")

        if "docstring" in subdict:
            self[key]["docstring"] = subdict.get("docstring").format(
                **self.get_defaults(key)
            )

    def get_defaults(self, key):
        """
        Get the default values for a given equation.

        Parameters
        ----------
        key: str
            The equation name.

        Returns
        -------
        defaults: dict
            A dictionary of default values.
        """

        defaults = {
            key: Quantity(item)._repr_latex_().replace("$", "")
            for key, item in self[key]["default_fiducial_values"].items()
        }

        defaults["name"] = key

        return defaults


#: equation definitions
EQN_DEFINITIONS = EqDict()

EQN_DEFINITIONS["h0"] = {
    "description": "Gravitational wave amplitude",
    "variable": "h0",
    "latex_string": "h_0",
    "default_fiducial_values": {
        "ellipticity": 1e-6,
        "momentofinertia": 1e38 * u.Unit("kg m^2"),
        "rotationfrequency": 100 * u.Hz,
        "distance": 1 * u.kpc,
    },
    "parts": [
        ("16", "1"),
        ("pi", "2"),
        ("G", "1"),
        ("c", "-4"),
        ("ellipticity", "1"),
        ("momentofinertia", "1"),
        ("rotationfrequency", "2"),
        ("distance", "-1"),
    ],
    "alternative_variables": ["gwfrequency", "rotationperiod"],
    "converters": {
        "rotationfrequency": convert_to_rotation_frequency,
    },
    "reference": {
        "short": "Jaranowski, P., Krolak, A., & Schutz, B. F. 1998, PhRvD, 58, 063001",
        "adsurl": "https://ui.adsabs.harvard.edu/abs/1998PhRvD..58f3001J/abstract",
        "eqno": "23",
        "bibtex": r"""\
@ARTICLE{1998PhRvD..58f3001J,
       author = {{Jaranowski}, Piotr and {Kr{\'o}lak}, Andrzej and {Schutz}, Bernard F.},
        title = "{Data analysis of gravitational-wave signals from spinning neutron stars: The signal and its detection}",
      journal = {\prd},
     keywords = {95.55.Ym, 04.80.Nn, 95.75.Pq, 97.60.Gb, Gravitational radiation detectors, mass spectrometers, and other instrumentation and techniques, Gravitational wave detectors and experiments, Mathematical procedures and computer techniques, Pulsars, General Relativity and Quantum Cosmology},
         year = 1998,
        month = sep,
       volume = {58},
       number = {6},
          eid = {063001},
        pages = {063001},
          doi = {10.1103/PhysRevD.58.063001},
archivePrefix = {arXiv},
       eprint = {gr-qc/9804014},
 primaryClass = {gr-qc},
       adsurl = {https://ui.adsabs.harvard.edu/abs/1998PhRvD..58f3001J},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    },
    "docstring": """
Generate the gravitational-wave amplitude for a signal emitted from the l=m=2
mass quadrupole mode.

For the optional input keyword parameters below a range of aliases, as given in
:obj:`~cweqgen.definitions.ALLOWED_VARIABLES`, can be used instead.

:param str equation: "{name}"
:keyword float or ~astropy.units.quantity.Quantity ellipticity: The ellipticity of the source with which the calculate the GW amplitude. The default value is :math:`{ellipticity}`.
:keyword float or ~astropy.units.quantity.Quantity momentofinertia: The principal moment of inertia with which the calculate the GW amplitude. If given as a float units of kg m^2 are assumed. The default value is :math:`{momentofinertia}`.
:keyword float or ~astropy.units.quantity.Quantity rotationfrequency: The rotation frequency of the source. If given as a float units of Hz are assumed. The default value is :math:`{rotationfrequency}`. If the rotational period or gravitational wave frequency are given instead then they will be converted into rotational frequency (for GW frequency it is assumed that this is twice the rotational frequency).
:keyword float or ~astropy.units.quantity.Quantity distance: The distance to the source. If given as a float units of kpc are assumed. The default value is :math:`{distance}`.
""",
}

EQN_DEFINITIONS["spindownluminosity"] = {
    "description": "The spin-down luminosity of a pulsar",
    "variable": "luminosity",
    "latex_string": r"L_{\rm sd}",
    "default_fiducial_values": {
        "momentofinertia": 1e38 * u.Unit("kg m^2"),
        "rotationfrequency": 100 * u.Hz,
        "rotationfdot": -1e-11 * u.Hz / u.s,
    },
    "parts": [
        ("4", "1"),
        ("pi", "2"),
        ("momentofinertia", "1"),
        ("rotationfrequency", "1"),
        ("rotationfdot", "1"),
    ],
    "alternative_variables": [
        "gwfrequency",
        "rotationperiod",
        "gwfdot",
        "rotationpdot",
    ],
    "converters": {
        "rotationfrequency": convert_to_rotation_frequency,
        "rotationfdot": convert_to_rotation_fdot,
    },
    "reference": {  # this is just an example reference for the braking index (there will be earlier references!)
        "short": "Condon, J. J. and Ransom, S. M., 2016, Essential Radio Astronomy",
        "adsurl": "https://ui.adsabs.harvard.edu/abs/2016era..book.....C/abstract",
        "eqno": "6.35",
        "bibtex": r"""\
@BOOK{2016era..book.....C,
       author = {{Condon}, James J. and {Ransom}, Scott M.},
        title = "{Essential Radio Astronomy}",
         year = 2016,
       adsurl = {https://ui.adsabs.harvard.edu/abs/2016era..book.....C},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    },
    "docstring": """
Generate the spin-down luminosity of a pulsar.

For the optional input keyword parameters below a range of aliases, as given in
:obj:`~cweqgen.definitions.ALLOWED_VARIABLES`, can be used instead.

:param str equation: "{name}"
:keyword float or ~astropy.units.quantity.Quantity momentofinertia: The principal moment of inertia with which the calculate the GW amplitude. If given as a float units of kg m^2 are assumed. The default value is :math:`{momentofinertia}`.
:keyword float or ~astropy.units.quantity.Quantity rotationfrequency: The rotation frequency of the source. If given as a float units of Hz are assumed. The default value is :math:`{rotationfrequency}`. If the rotational period or gravitational wave frequency are given instead then they will be converted into rotational frequency (for GW frequency it is assumed that this is twice the rotational frequency).
:keyword float or ~astropy.units.quantity.Quantity rotationfdot: The first rotational frequency derivative (i.e. the spin-down). If given as a float units of Hz/s are assumed. The default value is :math:`{rotationfdot}`. If the rotational period derivative or gravitational wave frequency derivative is given instead then they will be converted into rotational frequency derivative.
""",
}

EQN_DEFINITIONS["gwluminosity"] = {
    "description": "The gravitational-wave luminosity of a pulsar",
    "variable": "luminosity",
    "latex_string": r"L_{\rm gw}",
    "default_fiducial_values": {
        "momentofinertia": 1e38 * u.Unit("kg m^2"),
        "rotationfrequency": 100 * u.Hz,
        "ellipticity": 1e-6,
    },
    "parts": [
        ("2048/5", "1"),
        ("pi", "6"),
        ("G", "1"),
        ("c", "-5"),
        ("momentofinertia", "2"),
        ("ellipticity", "2"),
        ("rotationfrequency", "6"),
    ],
    "alternative_variables": [
        "gwfrequency",
        "rotationperiod",
    ],
    "converters": {
        "rotationfrequency": convert_to_rotation_frequency,
    },
    "reference": {
        "short": "Ostriker, J. P. and Gunn, J. E., 1969, ApJ, 157, 1395",
        "adsurl": "https://ui.adsabs.harvard.edu/abs/1969ApJ...157.1395O/abstract",
        "eqno": "7",
        "bibtex": r"""\
@ARTICLE{1969ApJ...157.1395O,
       author = {{Ostriker}, J.~P. and {Gunn}, J.~E.},
        title = "{On the Nature of Pulsars. I. Theory}",
      journal = {\apj},
         year = 1969,
        month = sep,
       volume = {157},
        pages = {1395},
          doi = {10.1086/150160},
       adsurl = {https://ui.adsabs.harvard.edu/abs/1969ApJ...157.1395O},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    },
    "docstring": """
Generate the gravitational-wave luminosity of a pulsar.

For the optional input keyword parameters below a range of aliases, as given in
:obj:`~cweqgen.definitions.ALLOWED_VARIABLES`, can be used instead.

:param str equation: "{name}"
:keyword float or ~astropy.units.quantity.Quantity ellipticity: The ellipticity of the source with which the calculate the GW amplitude. The default value is :math:`{ellipticity}`.
:keyword float or ~astropy.units.quantity.Quantity momentofinertia: The principal moment of inertia with which the calculate the GW amplitude. If given as a float units of kg m^2 are assumed. The default value is :math:`{momentofinertia}`.
:keyword float or ~astropy.units.quantity.Quantity rotationfrequency: The rotation frequency of the source. If given as a float units of Hz are assumed. The default value is :math:`{rotationfrequency}`. If the rotational period or gravitational wave frequency are given instead then they will be converted into rotational frequency (for GW frequency it is assumed that this is twice the rotational frequency).
""",
}

EQN_DEFINITIONS["h0spindown"] = {
    "description": "Gravitational-wave amplitude spin-down limit",
    "variable": "h0",
    "latex_string": r"h_0^{\rm sd}",
    "default_fiducial_values": {
        "momentofinertia": 1e38 * u.Unit("kg m^2"),
        "rotationfrequency": 100 * u.Hz,
        "rotationfdot": -1e-11 * u.Hz / u.s,
        "distance": 1 * u.kpc,
    },
    "chain": [
        "ellipticityspindown",
        "substitute h0",
    ],
    "alternative_variables": [
        "gwfrequency",
        "rotationperiod",
        "gwfdot",
        "rotationpdot",
    ],
    "converters": {
        "rotationfrequency": convert_to_rotation_frequency,
        "rotationfdot": convert_to_rotation_fdot,
    },
    "reference": {
        "short": "Aasi, A., et al. 2014, ApJ, 785, 119",
        "adsurl": "https://ui.adsabs.harvard.edu/abs/2014ApJ...785..119A/abstract",
        "eqno": "5",
        "bibtex": r"""\
@ARTICLE{2014ApJ...785..119A,
       author = {{Aasi}, J. and others},
       title = "{Gravitational Waves from Known Pulsars: Results from the Initial Detector Era}",
      journal = {\apj},
     keywords = {gravitational waves, pulsars: general, Astrophysics - High Energy Astrophysical Phenomena, General Relativity and Quantum Cosmology},
         year = 2014,
        month = apr,
       volume = {785},
       number = {2},
          eid = {119},
        pages = {119},
          doi = {10.1088/0004-637X/785/2/119},
archivePrefix = {arXiv},
       eprint = {1309.4027},
 primaryClass = {astro-ph.HE},
       adsurl = {https://ui.adsabs.harvard.edu/abs/2014ApJ...785..119A},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    },
    "docstring": """
Generate the equation for the spin-down limit on the gravitational wave
amplitude for a signal emitted from the l=m=2 mass quadrupole mode.

For the optional input keyword parameters below a range of aliases, as given in
:obj:`~cweqgen.definitions.ALLOWED_VARIABLES`, can be used instead.

:param str equation: "{name}"
:keyword float or ~astropy.units.quantity.Quantity momentofinertia: The principal moment of inertia with which the calculate the GW amplitude. If given as a float units of kg m^2 are assumed. The default value is :math:`{momentofinertia}`.
:keyword float or ~astropy.units.quantity.Quantity rotationfrequency: The rotation frequency of the source. If given as a float units of Hz are assumed. The default value is :math:`{rotationfrequency}`. If the rotational period or gravitational wave frequency are given instead then they will be converted into rotational frequency (for GW frequency it is assumed that this is twice the rotational frequency).
:keyword float or ~astropy.units.quantity.Quantity distance: The distance to the source. If given as a float units of kpc are assumed. The default value is :math:`{distance}`.
:keyword float or ~astropy.units.quantity.Quantity rotationfdot: The first rotational frequency derivative (i.e. the spin-down). If given as a float units of Hz/s are assumed. The default value is :math:`{rotationfdot}`. If the rotational period derivative or gravitational wave frequency derivative is given instead then they will be converted into rotational frequency derivative.
""",
}

EQN_DEFINITIONS["ellipticityspindown"] = {
    "description": "Spin-down limit for neutron star ellipticity",
    "variable": "ellipticity",
    "latex_string": r"\varepsilon^{\rm sd}",
    "default_fiducial_values": {
        "momentofinertia": 1e38 * u.Unit("kg m^2"),
        "rotationfrequency": 100 * u.Hz,
        "rotationfdot": -1e-11 * u.Hz / u.s,
    },
    "chain": [
        "gwluminosity",
        "equals spindownluminosity",
        "rearrange ellipticity",
    ],
    "alternative_variables": [
        "gwfrequency",
        "rotationperiod",
        "gwfdot",
        "rotationpdot",
    ],
    "converters": {
        "rotationfrequency": convert_to_rotation_frequency,
        "rotationfdot": convert_to_rotation_fdot,
    },
    "reference": {
        "short": "Abbott, B. P., et al. 2019, ApJ, 879, 10",
        "adsurl": "https://ui.adsabs.harvard.edu/abs/2019ApJ...879...10A/abstract",
        "eqno": "A9",
        "bibtex": r"""\
@ARTICLE{2019ApJ...879...10A,
       author = {{Aasi}, J. and others},
        title = "{Searches for Gravitational Waves from Known Pulsars at Two Harmonics in 2015-2017 LIGO Data}",
      journal = {\apj},
     keywords = {gravitational waves, pulsars: general, stars: neutron, Astrophysics - High Energy Astrophysical Phenomena, General Relativity and Quantum Cosmology},
         year = 2019,
        month = jul,
       volume = {879},
       number = {1},
          eid = {10},
        pages = {10},
          doi = {10.3847/1538-4357/ab20cb},
archivePrefix = {arXiv},
       eprint = {1902.08507},
 primaryClass = {astro-ph.HE},
       adsurl = {https://ui.adsabs.harvard.edu/abs/2019ApJ...879...10A},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    },
    "docstring": """
Generate the equation for the spin-down limit on the source ellipticity
for a signal emitted from the l=m=2 mass quadrupole mode.

For the optional input keyword parameters below a range of aliases, as
given in :obj:`cweqgen.definitions.ALLOWED_VARIABLES`, can be used instead.

:param str equation: "{name}"
:keyword float or ~astropy.units.quantity.Quantity momentofinertia: The principle moment of inertia with which the calculate the spin-down limit. If given as a float units of kg m^2 are assumed. The default value is :math:`{momentofinertia}`.
:keyword float or ~astropy.units.quantity.Quantity rotationfrequency: The rotation frequency of the source. If given as a float units of Hz are assumed. The default value is :math:`{rotationfrequency}`. If the rotational period or gravitational wave frequency are given instead then they will be converted into rotational frequency (for GW frequency it is assumed that this is twice the rotational frequency).
:keyword float or ~astropy.units.quantity.Quantity rotationfdot: The first rotational frequency derivative (i.e. the spin-down). If given as a float units of Hz/s are assumed. The default value is :math:`{rotationfdot}`. If the rotational period derivative or gravitational wave frequency derivative is given instead then they will be converted into rotational frequency derivative.
""",
}

EQN_DEFINITIONS["brakingindex"] = {
    "description": "The braking index of a pulsar",
    "variable": "brakingindex",
    "latex_string": "n",
    "default_fiducial_values": {
        "rotationfrequency": 50 * u.Hz,
        "rotationfddot": 1e-23 * u.Hz / (u.s ** 2),
        "rotationfdot": -1e-11 * u.Hz / u.s,
    },
    "parts": [
        ("rotationfrequency", "1"),
        ("rotationfddot", "1"),
        ("rotationfdot", "-2"),
    ],
    "alternative_variables": [
        "gwfrequency",
        "rotationperiod",
        "gwfdot",
        "rotationpdot",
    ],
    "converters": {
        "rotationfrequency": convert_to_rotation_frequency,
        "rotationfdot": convert_to_rotation_fdot,
    },
    "reference": {  # this is just an example reference for the braking index (there will be earlier references!)
        "short": "Condon, J. J. and Ransom, S. M., 2016, Essential Radio Astronomy",
        "adsurl": "https://ui.adsabs.harvard.edu/abs/2016era..book.....C/abstract",
        "eqno": "6.35",
        "bibtex": r"""\
@BOOK{2016era..book.....C,
       author = {{Condon}, James J. and {Ransom}, Scott M.},
        title = "{Essential Radio Astronomy}",
         year = 2016,
       adsurl = {https://ui.adsabs.harvard.edu/abs/2016era..book.....C},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    },
    "docstring": """
Generate the equation for the braking index of a pulsar.

For the optional input keyword parameters below a range of aliases, as
given in :obj:`cweqgen.definitions.ALLOWED_VARIABLES`, can be used instead.

:param str equation: "{name}"
:keyword float or ~astropy.units.quantity.Quantity rotationfrequency: The rotation frequency of the source. If given as a float units of Hz are assumed. The default value is :math:`{rotationfrequency}`. If the rotational period or gravitational wave frequency are given instead then they will be converted into rotational frequency (for GW frequency it is assumed that this is twice the rotational frequency).
:keyword float or ~astropy.units.quantity.Quantity rotationfdot: The first rotational frequency derivative (i.e. the spin-down). If given as a float units of Hz/s are assumed. The default value is :math:`{rotationfdot}`. If the rotational period derivative or gravitational wave frequency derivative is given instead then they will be converted into rotational frequency derivative.
:keyword float or ~astropy.units.quantity.Quantity rotationfddot: The second rotational frequency derivative. If given as a float units of Hz/s^2 are assumed. The default value is :math:`{rotationfddot}`.
""",
}

EQN_DEFINITIONS["characteristicage"] = {
    "description": "The characteristic age of a pulsar",
    "variable": "characteristicage",
    "latex_string": r"\tau",
    "default_fiducial_values": {
        "brakingindex": 3,
        "rotationperiod": 0.01 * u.s,
        "rotationpdot": 1e-15 * u.s / u.s,
    },
    "parts": [
        ("rotationperiod", "1"),
        ("rotationpdot", "-1"),
        ("brakingindex - 1", "-1"),
    ],
    "alternative_variables": ["gwfrequency", "rotationfrequency", "rotationfdot"],
    "converters": {
        "rotationperiod": convert_to_rotation_period,
        "rotationpdot": convert_to_rotation_pdot,
    },
    "reference": {
        "short": "Condon, J. J. and Ransom, S. M., 2016, Essential Radio Astronomy",
        "adsurl": "https://ui.adsabs.harvard.edu/abs/2016era..book.....C/abstract",
        "eqno": "6.31",
        "bibtex": r"""\
@BOOK{2016era..book.....C,
       author = {{Condon}, James J. and {Ransom}, Scott M.},
        title = "{Essential Radio Astronomy}",
         year = 2016,
       adsurl = {https://ui.adsabs.harvard.edu/abs/2016era..book.....C},
      adsnote = {Provided by the SAO/NASA Astrophysics Data System}
}""",
    },
    "docstring": """
Generate the pulsar's characteristic age.

For the optional input keyword parameters below a range of aliases, as given in
:obj:`~cweqgen.definitions.ALLOWED_VARIABLES`, can be used instead.

:param str equation: "{name}"
:keyword float or ~astropy.units.quantity.Quantity rotationperiod: The rotation period of a pulsar. If given as a float units of s are assumed. The default value is :math:`{rotationperiod}`.
:keyword float or ~astropy.units.quantity.Quantity rotationpdot: The first derivative of the rotation period of a pulsar. If given as a float units of s/s are assumed. The default value is :math:`{rotationpdot}`.
:keyword float or ~astropy.units.quantity.Quantity brakingindex: The braking index of the pulsar. The default value is :math:`{brakingindex}`.
""",
}


SUPPLEMENTAL_EQUATIONS = EqDict()

SUPPLEMENTAL_EQUATIONS["rotationfrequency_to_period"] = {
    "description": "The rotation frequency of the pulsar in terms of the rotation period",
    "variable": "rotationfrequency",
    "latex_string": r"f_{\rm rot}",
    "default_fiducial_values": {
        "rotationperiod": 0.01 * u.s,
    },
    "parts": [
        ("rotationperiod", "-1"),
    ],
}

SUPPLEMENTAL_EQUATIONS["gwfrequency_to_rotationfrequency"] = {
    "description": "The GW frequency of the pulsar in terms of the rotation frequency",
    "variable": "gwfrequency",
    "latex_string": r"f_{\rm gw}",
    "default_fiducial_values": {
        "rotationfrequency": 100 * u.Hz,
    },
    "parts": [
        ("2", "1"),
        ("rotationfrequency", "1"),
    ],
}

SUPPLEMENTAL_EQUATIONS["angulargwfrequency_to_gwfrequency"] = {
    "description": "The angular GW frequency of the pulsar in terms of the GW frequency",
    "variable": "angulargwfrequency",
    "latex_string": r"\Omega_{\rm gw}",
    "default_fiducial_values": {
        "gwfrequency": 200 * u.Hz,
    },
    "parts": [
        ("2", "1"),
        ("pi", "1"),
        ("gwfrequency", "1"),
    ],
}

SUPPLEMENTAL_EQUATIONS["angularrotationfrequency_to_angulargwfrequency"] = {
    "description": "The angular rotation frequency of the pulsar in terms of the angular GW frequency",
    "variable": "angularrotationfrequency",
    "latex_string": r"\Omega_{\rm rot}",
    "default_fiducial_values": {
        "angulargwfrequency": 2 * pi * 200 * u.rad / u.s,
    },
    "parts": [
        ("1/2", "1"),
        ("angulargwfrequency", "1"),
    ],
}

SUPPLEMENTAL_EQUATIONS["rotationperiod_to_angularrotationfrequency"] = {
    "description": "The rotation period of the pulsar in terms of the angular rotation frequency",
    "variable": "rotationperiod",
    "latex_string": "P",
    "default_fiducial_values": {
        "angularrotationfrequency": 2 * pi * 100 * u.rad / u.s,
    },
    "parts": [
        ("2", "1"),
        ("pi", "1"),
        ("angularrotationfrequency", "-1"),
    ],
}

SUPPLEMENTAL_EQUATIONS["rotationfdot_to_period"] = {
    "description": "The rotation frequency derivative in terms of period and period derivative",
    "variable": "rotationfdot",
    "latex_string": r"\dot{f}_{\rm rot}",
    "default_fiducial_values": {
        "rotationperiod": 0.01 * u.s,
        "rotationpdot": 1e-15 * u.s / u.s,
    },
    "parts": [
        ("-1", "1"),
        ("rotationpdot", "1"),
        ("rotationperiod", "-2"),
    ],
}

SUPPLEMENTAL_EQUATIONS["rotationpdot_to_angularrotationfdot"] = {
    "description": "The rotation period derivative in terms of angular rotation frequency and its derivative",
    "variable": "rotationpdot",
    "latex_string": r"\dot{P}",
    "default_fiducial_values": {
        "angularrotationfrequency": 2 * pi * 100 * u.rad / u.s,
        "angularrotationfdot": -2 * pi * 1e-11 * u.rad / (u.s ** 2),
    },
    "parts": [
        ("-2", "1"),
        ("pi", "1"),
        ("angularrotationfdot", "1"),
        ("angularrotationfrequency", "-2"),
    ],
}

SUPPLEMENTAL_EQUATIONS["angularrotationfdot_to_angulargwfdot"] = {
    "description": "The angular rotation frequency derivative in terms of angular gravitational-wave frequency derivative",
    "variable": "angularrotationfdot",
    "latex_string": r"\dot{\Omega}_{\rm rot}",
    "default_fiducial_values": {
        "angulargwfdot": 4 * pi * 100 * u.rad / (u.s ** 2),
    },
    "parts": [
        ("1/2", "1"),
        ("angulargwfdot", "1"),
    ],
}

SUPPLEMENTAL_EQUATIONS["angulargwfdot_to_gwfdot"] = {
    "description": "The angular gravitational-wave frequency derivative in terms of gravitational-wave frequency derivative",
    "variable": "angulargwfdot",
    "latex_string": r"\dot{\Omega}_{\rm gw}",
    "default_fiducial_values": {
        "gwfdot": 2 * 100 * u.Hz / u.s,
    },
    "parts": [
        ("2", "1"),
        ("pi", "1"),
        ("gwfdot", "1"),
    ],
}

SUPPLEMENTAL_EQUATIONS["gwfdot_to_rotationfdot"] = {
    "description": "The angular rotation frequency derivative in terms of angular gravitational-wave frequency derivative",
    "variable": "gwfdot",
    "latex_string": r"\dot{f}_{\rm gw}",
    "default_fiducial_values": {
        "rotationfdot": 100 * u.Hz / u.s,
    },
    "parts": [
        ("2", "1"),
        ("rotationfdot", "1"),
    ],
}
